#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "openglheader.h"   /* najpierw ten */
#include <GL/freeglut.h>    /* potem ten */

#include "utilities.h"
#include "lights.h"
#include "mygltext.h"
#include "readsmf.h"
#include "app1e.h"

GLuint  shader_id[5];
GLuint  program_id[2];
GLuint  trbi, trbuf, trbbp;
GLint   trbsize, trbofs[5];

int     smf_nv, smf_ntr;
GLfloat smf_mm[16];
GLuint  smf_vao, smf_vbo[2];

float   model_rot_axis[3] = {0.0,1.0,0.0},
        model_rot_angle0 = 0.0, model_rot_angle;
const float viewer_pos0[4] = {0.0,0.0,10.0,1.0};
TransBl trans;               

myTextObject *vptext;
myFont  *font;

void LoadMyShaders ( void )
{
  static const char *filename[] =
    { "app1b0.glsl.vert", "app1b0.glsl.frag",
      "app1b1.glsl.vert", "app1b1.glsl.geom", "app1b1.glsl.frag" };
  static const GLchar *UTBNames[] =
    { "TransBlock", "TransBlock.mm", "TransBlock.vm", "TransBlock.pm",
      "TransBlock.mvpm", "TransBlock.eyepos" };
  static const GLchar *ULSNames[] =
    { "LSBlock", "LSBlock.nls", "LSBlock.mask",
      "LSBlock.ls[0].ambient", "LSBlock.ls[0].direct", "LSBlock.ls[0].position",
      "LSBlock.ls[0].attenuation", "LSBlock.ls[1].ambient" };

  shader_id[0] = CompileShaderFiles ( GL_VERTEX_SHADER, 1, &filename[0] );
  shader_id[1] = CompileShaderFiles ( GL_FRAGMENT_SHADER, 1, &filename[1]);
  program_id[0] = LinkShaderProgram ( 2, &shader_id[0] );
  shader_id[2] = CompileShaderFiles ( GL_VERTEX_SHADER, 1, &filename[2] );
  shader_id[3] = CompileShaderFiles ( GL_GEOMETRY_SHADER, 1, &filename[3] );
  shader_id[4] = CompileShaderFiles ( GL_FRAGMENT_SHADER, 1, &filename[4] );
  program_id[1] = LinkShaderProgram ( 3, &shader_id[2] );
  GetAccessToUniformBlock ( program_id[1], 5, &UTBNames[0],
                            &trbi, &trbsize, trbofs, &trbbp );
  GetAccessToUniformBlock ( program_id[1], 7, &ULSNames[0],
                            &lsbi, &lsbsize, lsbofs, &lsbbp );
  AttachUniformBlockToBP ( program_id[0], UTBNames[0], trbbp );
  trbuf = NewUniformBlockObject ( trbsize, trbbp );
  lsbuf = NewUniformBlockObject ( lsbsize, lsbbp );
  ExitIfGLError ( "LoadMyShaders" );
} /*LoadMyShaders*/

void SetupMVPMatrix ( void )
{
  GLfloat m[16], mvp[16];

  M4x4Multf ( m, trans.vm, trans.mm );
  M4x4Multf ( mvp, trans.pm, m );
  glBindBuffer ( GL_UNIFORM_BUFFER, trbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[3], 16*sizeof(GLfloat), mvp );
  ExitIfGLError ( "SetupNVPMatrix" );
} /*SetupMVPMatrix*/

void SetupModelMatrix ( float axis[3], float angle )
{
  GLfloat mm[16];

  M4x4RotateVf ( mm, axis[0], axis[1], axis[2], angle );
  M4x4Multf ( trans.mm, mm, smf_mm );
  glBindBuffer ( GL_UNIFORM_BUFFER, trbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[0], 16*sizeof(GLfloat), trans.mm );
  ExitIfGLError ( "SetupModelMatrix" );
  SetupMVPMatrix ();
} /*SetupModelMatrix*/

void InitViewMatrix ( void )
{
  memcpy ( trans.eyepos, viewer_pos0, 4*sizeof(GLfloat) );
  M4x4Translatef ( trans.vm, -viewer_pos0[0], -viewer_pos0[1], -viewer_pos0[2] );
  glBindBuffer ( GL_UNIFORM_BUFFER, trbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[1], 16*sizeof(GLfloat), trans.vm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[4], 4*sizeof(GLfloat), trans.eyepos );
  ExitIfGLError ( "InitViewMatrix" );
  SetupMVPMatrix ();
} /*InitViewMatrix*/

void RotateViewer ( int delta_xi, int delta_eta )
{
  float   vi[3], lgt, angi, vk[3], angk;
  GLfloat tm[16];

  if ( delta_xi == 0 && delta_eta == 0 )
    return;  /* natychmiast uciekamy - nie chcemy dzielic przez 0 */
  vi[0] = (float)delta_eta*(right-left)/(float)win_height;
  vi[1] = (float)delta_xi*(top-bottom)/(float)win_width;
  vi[2] = 0.0;
  lgt = sqrt ( V3DotProductf ( vi, vi ) );
  vi[0] /= lgt;
  vi[1] /= lgt;
  angi = -0.052359878;  /* -3 stopnie */
  V3CompRotationsf ( vk, &angk, viewer_rvec, viewer_rangle, vi, angi );
  memcpy ( viewer_rvec, vk, 3*sizeof(float) );
  viewer_rangle = angk;
  M4x4Translatef ( trans.vm, -viewer_pos0[0], -viewer_pos0[1], -viewer_pos0[2] );
  M4x4MRotateVf ( trans.vm, viewer_rvec[0], viewer_rvec[1], viewer_rvec[2],
                  -viewer_rangle );
  glBindBuffer ( GL_UNIFORM_BUFFER, trbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[1], 16*sizeof(GLfloat), trans.vm );
  M4x4Transposef ( tm, trans.vm );  tm[3] = tm[7] = tm[11] = 0.0;
  M4x4MultMVf ( trans.eyepos, tm, viewer_pos0 );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[4], 4*sizeof(GLfloat), trans.eyepos );
  ExitIfGLError ( "RotateViewer" );
  SetupMVPMatrix ();
} /*RotateViewer*/

void FindSMFMMatrix ( int smf_nv, GLfloat *vc, GLfloat smf_mm[16] )
{
  GLfloat xmin, xmax, ymin, ymax, zmin, zmax, dx, dy, dz, s, sm[16], tm[16];
  int     i, j;

  if ( smf_nv > 1 ) {
    xmin = xmax = vc[0];
    ymin = ymax = vc[1];
    zmin = zmax = vc[2];
    for ( i = 1, j = 3;  i < smf_nv;  i++, j += 3 ) {
      if ( vc[j] < xmin )      xmin = vc[j];
      else if ( vc[j] > xmax ) xmax = vc[j];
      if ( vc[j+1] < ymin )      ymin = vc[j+1];
      else if ( vc[j+1] > ymax ) ymax = vc[j+1];
      if ( vc[j+2] < zmin )      zmin = vc[j+2];
      else if ( vc[j+2] > zmax ) zmax = vc[j+2];
    }
    dx = 0.5*(xmax-xmin);
    dy = 0.5*(ymax-ymin);
    dz = 0.5*(zmax-zmin);
    s = dx > dy ? dx : dy;
    s = dz > s  ? dz : s;
    if ( s > 0.0 ) {
      s = 1.0/s;  M4x4Scalef ( sm, s, s, s );
      M4x4Translatef ( tm, -0.5*(xmin+xmax), -0.5*(ymin+ymax), -0.5*(zmin+zmax) );
      M4x4Multf ( smf_mm, sm, tm );
      return;
    }
  }
  M4x4Identf ( smf_mm );
} /*FindSMFMMatrix*/

void ConstructSMFObject ( char *fn )
{
  GLfloat *vc;
  GLint   *trv;

  if ( ReadSMFFile ( fn, &smf_nv, &vc, &smf_ntr, &trv ) ) {
    glGenVertexArrays ( 1, &smf_vao );
    glBindVertexArray ( smf_vao );
    glGenBuffers ( 2, smf_vbo );
    glBindBuffer ( GL_ARRAY_BUFFER, smf_vbo[0] );
    glBufferData ( GL_ARRAY_BUFFER, smf_nv*3*sizeof(GLfloat),
                   vc, GL_STATIC_DRAW );
    glEnableVertexAttribArray ( 0 );
    glVertexAttrib4f ( 1, 0.5, 0.8, 1.0, 1.0 );
    glVertexAttribPointer ( 0, 3, GL_FLOAT, GL_FALSE,
                            3*sizeof(GLfloat), (GLvoid*)0 );
    glBindBuffer ( GL_ELEMENT_ARRAY_BUFFER, smf_vbo[1] );
    glBufferData ( GL_ELEMENT_ARRAY_BUFFER, 3*smf_ntr*sizeof(GLint),
                   trv, GL_STATIC_DRAW );
    FindSMFMMatrix ( smf_nv, vc, smf_mm );
    free ( vc );
    free ( trv );
    ExitIfGLError ( "ConstructSMFObject" );
  }
  else
    ExitOnError ( "Pliku nie udalo sie przeczytac" );
} /*ConstructSMFObject*/
 
void DrawSMFObject ( void )
{
  glBindVertexArray ( smf_vao );
  glBindBuffer ( GL_ELEMENT_ARRAY_BUFFER, smf_vbo[1] );
  glUseProgram ( program_id[enlight ? 1 : 0] );
  glDrawElements ( GL_TRIANGLES, 3*smf_ntr, GL_UNSIGNED_INT, 0 );
  ExitIfGLError ( "DrawSMFObject" );
} /*DrawSMFObject*/

void InitLights ( void )
{
  GLfloat amb0[4] = { 0.2, 0.2, 0.3, 1.0 };
  GLfloat dif0[4] = { 0.8, 0.8, 0.8, 1.0 };
  GLfloat pos0[4] = { 0.0, 1.0, 1.0, 0.0 };
  GLfloat atn0[3] = { 1.0, 0.0, 0.0 };

  SetLightAmbient ( 0, amb0 );
  SetLightDiffuse ( 0, dif0 );
  SetLightPosition ( 0, pos0 );
  SetLightAttenuation ( 0, atn0 );
  SetLightOnOff ( 0, 1 );
} /*InitLights*/

void NotifyViewerPos ( void )
{
  GLchar s[60];

  sprintf ( s, "x = %5.2f, y = %5.2f, z = %5.2f",
            trans.eyepos[0], trans.eyepos[1], trans.eyepos[2] );
  SetTextObjectContents ( vptext, s, 0, 17, font );
} /*NotifyViewerPos*/

void InitMyObject ( char *fn )
{
  TimerInit ();
  font = NewFont18x10 ();
/*  font = NewFont12x6 ();*/
  vptext = NewTextObject ( 60 );
  memset ( &trans, 0, sizeof(TransBl) );
  memset ( &light, 0, sizeof(LightBl) );
  ConstructSMFObject ( fn );
  SetupModelMatrix ( model_rot_axis, model_rot_angle );
  InitViewMatrix ();
  NotifyViewerPos ();
  InitLights ();
} /*InitMyObject*/

void Cleanup ( void )
{
  int i;

  glUseProgram ( 0 );
  for ( i = 0; i < 5; i++ )
    glDeleteShader ( shader_id[i] );
  glDeleteProgram ( program_id[0] );
  glDeleteProgram ( program_id[1] );
  glDeleteBuffers ( 1, &trbuf );
  glDeleteBuffers ( 1, &lsbuf );
  glDeleteVertexArrays ( 1, &smf_vao );
  glDeleteBuffers ( 2, smf_vbo );
  DeleteTextObject ( vptext );
  DeleteFontObject ( font );
  ExitIfGLError ( "Cleanup" );
  glutDestroyWindow ( WindowHandle );
} /*Cleanup*/

