#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "openglheader.h"
#include <GL/glx.h>

#include "utilities.h"
#include "xvariables.h"
#include "xwidgets.h"
#include "initglxctx.h"
#include "meshes.h"
#include "app3_a.h"
#include "app3_b.h"

Window      window[3];
GC          xgc;
char        terminate;
Atom        aAnimate;

int         window0_width, window0_height;
xwinmenu    *wm1, *wm2;
xwidget     *mywdg;
myglwindata myglwdata;
char        opti = 0;

char    meshsw[NPALMMESHES] = {true,false,false,false,false};

char str_EXIT[] = "Exit";

void Win1Callback ( struct xwidget *wdg, int msg, int key, int x, int y )
{
  switch ( msg ) {
case WDGMSG_BUTTON_PRESS:
    switch ( wdg->id ) {
  case BTN_ID_EXIT:
      terminate = true;
      break;
  default:
      break;
    }
    break;

case WDGMSG_SWITCH_CHANGE:
    switch ( wdg->id ) {
  case SW_ID_MESH1:  case SW_ID_MESH2:  case SW_ID_MESH3:  case SW_ID_MESH4:
      if ( meshsw[wdg->id-SW_ID_MESH0] ) {
        memset ( &meshsw[1], false, NPALMMESHES-1 );
        meshsw[wdg->id-SW_ID_MESH0] = true;
      }
      goto win2changed;
  case SW_ID_MESH0:
win2changed:
      wm2->changed = true;
      PostMenuExposeEvent ( wm2 );
      break;
  default:
      break;
    }

default:
    break;
  }
} /*Win1CallBack*/

xwinmenu *SetupApp3Menu ( void )
{
  xwinmenu *wm;
  int      i;

  if ( !(wm = NewWinMenu ( window[1], MENU_WIDTH, WIN0_HEIGHT, 0, 0,
                           NULL, NULL, Win1Callback )) )
    ExitOnError ( "SetupApp3Menu" );
  NewButton ( wm, BTN_ID_EXIT, 60, 18, 2, 2, str_EXIT );
  for ( i = 0; i < NPALMMESHES; i++ )
    NewSwitch ( wm, SW_ID_MESH0+i, 16, 16, 2+20*i, 22, NULL, &meshsw[i] );
  return wm;
} /*SetupApp3Menu*/

/* ////////////////////////////////////////////////////////////////////////// */
void Win0ConfigureNotify ( int width, int height )
{
  window0_width  = width;
  window0_height = height;
  XMoveResizeWindow ( xdisplay, window[1], 0, 0, MENU_WIDTH, height );
  XMoveResizeWindow ( xdisplay, window[2], MENU_WIDTH, 0, width-MENU_WIDTH, height );
} /*Win0ConfigureNotify*/

void Win0MessageProc ( XEvent *ev )
{
  switch ( ev->xany.type ) {
case ConfigureNotify:
    Win0ConfigureNotify ( ev->xconfigure.width, ev->xconfigure.height );
    break;
default:
    break;
  }
} /*Win0MessageProc*/

/* ////////////////////////////////////////////////////////////////////////// */
void AdvanceAnimation ( myglwindata *wd )
{
  if ( wd->animation ) {
    myglwdata.model_rot_angle = myglwdata.model_rot_angle0 +
                                myglwdata.speed * TimerToc ();
    SetupModelMatrix ( wd, model_rot_axis, myglwdata.model_rot_angle );
    wm2->changed = true;
    PostMenuExposeEvent ( wm2 );
    PostClientMessageEvent ( window[2], aAnimate, 8, NULL );
  }
} /*AdvanceAnimation*/

void ToggleAnimation ( myglwindata *wd, char on )
{
  if ( (wd->animation = on) ) {
    TimerTic ();
    PostClientMessageEvent ( window[2], aAnimate, 8, NULL );
  }
  else
    myglwdata.model_rot_angle0 = myglwdata.model_rot_angle;
  wm1->changed = true;
  PostExposeEvent ( window[0], window0_width, window0_height );
} /*ToggleAnimation*/

void My3DWidgetRedraw ( struct xwidget *wdg )
{
  glViewport ( 0, 0, wdg->/*wm->*/r.width, wdg->/*wm->*/r.height );
  DrawMyScene ();
} /*My3DWidgetRedraw*/

void My3DWidgetReshape ( xwidget *wdg, short w, short h )
{
  myglwindata *wd;

  wdg->r.width = w;  wdg->r.height = h;
  wd = (myglwindata*)wdg->data0;
  wd->bottom = -(wd->top = 0.5533);
  wd->left = -(wd->right = 0.5533*(float)w/(float)h );
  wd->near = 5.0;  wd->far = 15.0;
  wd->rl = wd->right - wd->left;  wd->tb = wd->top - wd->bottom;
  M4x4Frustumf ( wd->pm, NULL, wd->left, wd->right, wd->bottom, wd->top,
                 wd->near, wd->far );
  SetVPMatrix ( wd->vm, wd->pm, wd->eyepos );
} /*My3DWidgetReshape*/

char My3DWidgetInput ( struct xwidget *wdg, int msg, int key, int x, int y )
{
  myglwindata *wd;

  wd = (myglwindata*)wdg->data0;
  switch ( wdg->state ) {
case WDGSTATE_DEFAULT:
    switch ( msg ) {
  case XWMSG_BUTTON_PRESS:
      if ( key == Button1 ) {
        wd->prevx = x;  wd->prevy = y;
        wdg->state = WDGSTATE_VIEW_TURNING;
        GrabInput ( wdg );
        return true;
      }
      break;
  case XWMSG_KEY_PRESS:
      goto process_key;
  case WDGMSG_RECONFIGURE:
      My3DWidgetReshape ( wdg, x, y );
      break;
  case XWMSG_CLIENT_MESSAGE:
      goto process_client_message;
  default:
      break;
    }
    break;

case WDGSTATE_VIEW_TURNING:
    switch ( msg ) {
  case XWMSG_MOUSE_MOTION:
      if ( ((XMotionEvent*)wdg->wm->ev)->state & Button1Mask ) {
        if ( RotateViewer ( wd, (float)wdg->r.width, (float)wdg->r.height,
                            x - wd->prevx, y - wd->prevy ) ) {
          wd->prevx = x;  wd->prevy = y;
          wdg->wm->changed = true;
        }
      }
      else
        goto release;
      break;
  case XWMSG_BUTTON_RELEASE:
      if ( key == Button1 ) {
release:
        wdg->state = WDGSTATE_DEFAULT;
        UngrabInput ( wdg );
        return true;
      }
      break;
  case XWMSG_KEY_PRESS:
process_key:
      switch ( key ) {
    case ' ':
        ToggleAnimation ( wd, !wd->animation );
        return true;
    default:
        break;
      }
      break;
  case XWMSG_CLIENT_MESSAGE:
process_client_message:
      if ( ((XClientMessageEvent*)wdg->wm->ev)->message_type == aAnimate )
        AdvanceAnimation ( wd );
  default:
      break;
    }
    break;

default:
    break;
  }
  return false;
} /*My3DWidgetInput*/

xwidget *New3DWidget ( struct xwinmenu *wm, int id, int w, int h )
{
  return NewWidget ( wm, sizeof(xwidget), id, w, h, 0, 0,
                     My3DWidgetInput, My3DWidgetRedraw, &myglwdata, NULL );
} /*New3DWidget*/

void RedrawWin2 ( xwinmenu *wm )
{
  xwidget *wdg;

  if ( opti > 0 ) {
    opti --;
    PostExposeEvent ( wm->window, wm->r.width, wm->r.height );
  }
  else {
    for ( wdg = wm->wlist.next;  wdg;  wdg = wdg->link.next )
      wdg->redraw ( wdg );
    glFlush ();
  }
  glXSwapBuffers ( xdisplay, wm->window );
} /*RedrawWin2*/

void Win2Reshape ( xwinmenu *wm, short w, short h )
{
  xwidget *wdg;

  for ( wdg = wm->wlist.next;  wdg;  wdg = wdg->link.next )
    wdg->input ( wdg, WDGMSG_RECONFIGURE, 0, w, h );
  opti = 3;
  PostExposeEvent ( wm->window, w, h );
  ExitIfGLError ( "Win2Reshape" );
} /*Win2Reshape*/

void Win2Callback ( struct xwidget *wdg, int msg, int key, int x, int y )
{
  switch ( msg ) {
case WDGMSG_RECONFIGURE:
    Win2Reshape ( wdg->wm, x, y );
    break;
case XWMSG_CLIENT_MESSAGE:
    mywdg->input ( mywdg, msg, key, x, y );
    break;
default:
    break;
  }
} /*Win2Callback*/

xwinmenu *SetupApp3GLWindow ( void )
{
  xwinmenu *wm;

  if ( !(wm = NewWinMenu ( window[2], WIN0_WIDTH-MENU_WIDTH, WIN0_HEIGHT, 0, 0,
                           NULL, RedrawWin2, Win2Callback )) )
    ExitOnError ( "SetupApp3GLWindow 0" );
  if ( !(mywdg = New3DWidget ( wm, GLWIN_ID_VIEW, wm->r.width, wm->r.height )) )
    ExitOnError ( "SetupApp3GLWindow 1" );
  return wm;
} /*SetupApp3GLWindow*/

/* ////////////////////////////////////////////////////////////////////////// */
void MessageLoop ( void )
{
  XEvent ev;

  terminate = false;
  do {
    XNextEvent ( xdisplay, &ev );
    if ( ev.xany.window == window[0] )
      Win0MessageProc ( &ev );
    else if ( ev.xany.window == window[1] )
      WinMenuInput ( wm1, &ev );
    else if ( ev.xany.window == window[2] )
      WinMenuInput ( wm2, &ev );
  } while ( !terminate );
} /*MessageLoop*/

void InitApp3Windows ( int argc, char **argv, int major, int minor )
{
  int vattr[] =
    { GLX_RGBA,
      GLX_DOUBLEBUFFER,
      GLX_RED_SIZE,    8,
      GLX_GREEN_SIZE,  8,
      GLX_BLUE_SIZE,   8,
      GLX_DEPTH_SIZE, 24,
      GLX_SAMPLE_BUFFERS, True,
      GLX_SAMPLES,     8,
      None };
  XSetWindowAttributes swa;
  static const int     wx[3] = { 0, 0, MENU_WIDTH };
  static const int     wh[3] = { WIN0_WIDTH, MENU_WIDTH, WIN0_WIDTH-MENU_WIDTH };
  Window               upw;
  int                  i;

  if ( !(xdisplay = XOpenDisplay ( "" )) )
    ExitOnError ( "InitApp3Windows 0" );
  xscreen = DefaultScreen ( xdisplay );
  xrootwin = RootWindow ( xdisplay, xscreen );
  InitGLXContext ( major, minor, vattr );
  if ( !(xcolormap = XCreateColormap ( xdisplay, xrootwin,
                                       xvisual, AllocNone )) )
    ExitOnError ( "InitApp3Windows 1" );
  swa.colormap = xcolormap;
/*  swa.backing_store = WhenMapped;*/
  swa.event_mask = ExposureMask | StructureNotifyMask| ButtonPressMask |
                   ButtonReleaseMask | PointerMotionMask | KeyPressMask ;
  for ( i = 0, upw = xrootwin; i < 3; i++, upw = window[0] ) {
    window[i] = XCreateWindow ( xdisplay, upw, wx[i], 0, wh[i], WIN0_HEIGHT,
                             0, 24, InputOutput, xvisual,
                             CWColormap | /*CWBackingStore |*/ CWEventMask, &swa );
    XMapWindow ( xdisplay, window[i] );
  }
  XStoreName ( xdisplay, window[0], "Trzecia aplikacja" );
  xgc = XCreateGC ( xdisplay, window[1], 0, 0 );
  InitRGBXColourmap (); 
  InitWinMenuPalette ();
  if ( !glXMakeCurrent ( xdisplay, window[2], glxcontext ) )
    ExitOnError ( "InitApp3Windows 2" );
  GetGLProcAddresses ();
  window0_width = WIN0_WIDTH;  
  window0_height = WIN0_HEIGHT;
} /*InitApp3Windows*/

void Initialise ( int argc, char **argv )
{
  InitApp3Windows ( argc, argv, APP3_MAJOR, APP3_MINOR );
  aAnimate = XInternAtom ( xdisplay, "aAnimate", False );
  wm1 = SetupApp3Menu ();
  wm2 = SetupApp3GLWindow ();
  LoadMyShaders ();
  InitMyObject ();
  Win2Reshape ( wm2, wm2->r.width, wm2->r.height );
} /*Initialise*/

void Cleanup ( void )
{
  DestroyMyObject ();
  DestroyMyShaders ();
  DestroyWinMenu ( wm1 );
  DestroyWinMenu ( wm2 );
  XDestroySubwindows ( xdisplay, window[0] );
  XDestroyWindow ( xdisplay, window[0] );
  XFreeGC ( xdisplay, xgc );
  XFree ( glxfbc );  
  XCloseDisplay ( xdisplay );
} /*Cleanup*/

int main ( int argc, char **argv )
{
  Initialise ( argc, argv );
  MessageLoop ();
  Cleanup ();
  exit ( 0 );
} /*main*/
