#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "openglheader.h"
#include <GLFW/glfw3.h>

#include "utilities.h"
#include "linkage.h"
#include "meshes.h"
#include "palm.h"
#include "app3b_a.h"

GLuint       lktrbuf[2];
static float palmartprange[NKLARTPARAMS][2] =
  {{0.0,-0.5*PI},{0.0,-0.5},{0.0,0.4*PI},{0.0,0.4*PI},
   {0.025*PI,-0.1*PI},{0.0,0.5*PI},{0.0,0.5*PI},{0.0,0.5*PI},
   {0.025*PI,-0.025*PI},{0.0,0.5*PI},{0.0,0.5*PI},{0.0,0.5*PI},
   {0.025*PI,-0.025*PI},{0.0,0.5*PI},{0.0,0.5*PI},{0.0,0.5*PI},
   {0.12*PI,-0.03*PI},{0.0,0.5*PI},{0.0,0.5*PI},{0.0,0.5*PI}};

static GLuint art_shader_id, art_program_id;
static GLuint cpi, cpbp, ctrbi, ctrbp, ctribi, ctribp, dim_loc, ncp_loc, trnum_loc;
static GLint  cpofs[1];

void LoadLinkageArticulationShaders ( void )
{
  static const char *filename[] = { "app2h.glsl.comp" };
  static const GLchar *UCPNames[] = { "CPoints", "CPoints.cp" };
  static const GLchar *UCompTrNames[] =  { "Tr", "Tr.tr" };
  static const GLchar *UCompTrIndNames[] = { "TrInd", "TrInd.ind" };
  static const GLchar DimName[] = "dim";
  static const GLchar NcpName[] = "ncp";
  static const GLchar TrNumName[] = "trnum";
  GLint i;

  art_shader_id = CompileShaderFiles ( GL_COMPUTE_SHADER, 1, &filename[0] );
  art_program_id = LinkShaderProgram ( 1, &art_shader_id );
  GetAccessToUniformBlock ( art_program_id, 1, &UCPNames[0], &cpi, &i, cpofs, &cpbp );
  GetAccessToUniformBlock ( art_program_id, 1, &UCompTrNames[0],
                            &ctrbi, &i, &i, &ctrbp );
  GetAccessToUniformBlock ( art_program_id, 1, &UCompTrIndNames[0],
                            &ctribi, &i, &i, &ctribp );
  dim_loc = glGetUniformLocation ( art_program_id, DimName );
  ncp_loc = glGetUniformLocation ( art_program_id, NcpName );
  trnum_loc = glGetUniformLocation ( art_program_id, TrNumName );
} /*LoadLinkageArticulationShaders*/

void DestroyLinkageArticulationShaders ( void )
{
  glUseProgram ( 0 );
  glDeleteShader ( art_shader_id );
  glDeleteProgram ( art_program_id );
} /*DestroyLinkageArticulationShaders*/

static char KLInitPalmMesh ( kl_object *obj, void *objdata, void *lkgdata )
{
  GPUmesh **palms;
  int     i;

  palms = (GPUmesh**)objdata;
  palms[0] = EnterPalmToGPU ();
  for ( i = 1; i <= NPALMMESHES+1; i++ ) {
    if ( !(palms[i] = malloc ( sizeof(GPUmesh) )) )
      ExitOnError ( "KLInitPalmMesh" );
    memset ( palms[i], 0, sizeof(GPUmesh) );
  }
  memcpy ( palms[1], palms[0], sizeof(GPUmesh) );
  glGenBuffers ( 1, &palms[1]->mbuf[2] );
  glBindBuffer ( GL_SHADER_STORAGE_BUFFER, palms[1]->mbuf[2] );
  glBufferData ( GL_SHADER_STORAGE_BUFFER, PALM_NV*3*sizeof(GLfloat),
                 NULL, GL_DYNAMIC_DRAW );
  return true;
} /*KLInitPalmMesh*/

static void KLDestroyPalmMesh ( kl_object *obj, void *lkgdata )
{
  GPUmesh **palms;
  int     i;

  palms = (GPUmesh**)obj->usrdata;
  for ( i = 0; i <= NPALMMESHES+1; i++ )
    DeleteGPUmesh ( palms[i] );
} /*KLDestroyPalmMesh*/

static void KLTransformVertices ( kl_object *obj, int refn, GLfloat *tr,
                                  int nv, int *vn, void *lkgdata )
{
  glBindBuffer ( GL_UNIFORM_BUFFER, lktrbuf[0] );
  glBufferSubData ( GL_UNIFORM_BUFFER, refn*16*sizeof(GLfloat),
                    16*sizeof(GLfloat), tr );
  ExitIfGLError ( "KLTransformVertices" );
} /*KLTransformVertices*/

static void KLPostprocessMesh ( kl_object *obj, void *lkgdata )
{
  mypalmscene *scene;
  GPUmesh     **mesh;
  int         i;

  scene = (mypalmscene*)lkgdata;
  glUseProgram ( art_program_id );
  mesh = (GPUmesh**)obj->usrdata;
  glBindBufferBase ( GL_UNIFORM_BUFFER, ctrbp, lktrbuf[0] );
  glBindBufferBase ( GL_UNIFORM_BUFFER, ctribp, lktrbuf[1] );
  glBindBufferBase ( GL_UNIFORM_BUFFER, cpbp, mesh[0]->mbuf[2] );
  glBindBufferBase ( GL_SHADER_STORAGE_BUFFER, 0, mesh[1]->mbuf[2] );
  glUniform1i ( dim_loc, obj->nvc );
  glUniform1i ( trnum_loc, -1 );
  glUniform1i ( ncp_loc, (GLint)obj->nvert );
  glDispatchCompute ( (obj->nvert+127)/128, 1, 1 );
  glMemoryBarrier ( GL_UNIFORM_BARRIER_BIT );
  if ( scene->lod >= 1 ) {
    for ( i = 1; i <= scene->lod; i++ ) {
      if ( !GPUmeshRefinement ( MESHDEG, mesh[i], mesh[i+1] ) )
        ExitOnError ( "KLPostprocessMesh" );
/*      printf ( "%d: nv = %d, nhe = %d, nfac = %d\n",
               i, mesh[i+1]->nv, mesh[i+1]->nhe, mesh[i+1]->nfac );*/
    }
    ComputeMeshNormalVectors ( mesh[scene->lod+1], 6, 3 );
  }
  ExitIfGLError ( "KLPostprocessMesh" );
} /*KLPostprocessMesh*/

static void KLRedrawMesh ( kl_object *obj, void *lkgdata )
{
  GLfloat edges_colour[4] = {0.0,0.5,0.7,1.0};
  GLfloat palm_colour[4] = {0.91,0.65,0.5,1.0};
  mypalmscene *scene;

  scene = (mypalmscene*)lkgdata;
  if ( scene->meshsw[0] )
    DrawMeshEdges ( scene->palms[1], edges_colour );
  if ( scene->lod >= 1 )
    DrawMeshFacets ( scene->palms[scene->lod+1], palm_colour, true );
} /*KLRedrawMesh*/

static void ConstructPalmMeshObj ( mypalmscene *scene,
                                   kl_linkage *lkg, int *obj, int *lnk )
{
  static int r0[19] = {0,1,2,3,6,7,8,9,10,13,16,17,135,138,139,140,141,142,143};
  static int r1[5] = {4,11,134,136,137};
  static int r2[8] = {5,12,14,15,18,19,20,21};
  static int r3[8] = {22,23,24,25,26,27,28,29};
  static int r4[8] = {30,31,32,33,34,35,36,37};
  static int r5[8] = {38,39,40,41,42,43,44,45};
  static int r6[8] = {46,47,48,49,50,51,52,53};
  static int r7[8] = {54,55,56,57,58,59,60,61};
  static int r8[8] = {62,63,64,65,66,67,68,69};
  static int r9[8] = {70,71,72,73,74,75,76,77};
  static int r10[8] = {78,79,80,81,82,83,84,85};
  static int r11[8] = {86,87,88,89,90,91,92,93};
  static int r12[8] = {94,95,96,97,98,99,100,101};
  static int r13[8] = {102,103,104,105,106,107,108,109};
  static int r14[8] = {110,111,112,113,114,115,116,117};
  static int r15[8] = {118,119,120,121,122,123,124,125};
  static int r16[8] = {126,127,128,129,130,131,132,133};

  obj[0] = kl_NewObject ( lkg, 3, PALM_NV, NULL, (void*)scene->palms,
                          KLInitPalmMesh, KLTransformVertices,
                          KLPostprocessMesh, KLRedrawMesh, KLDestroyPalmMesh );
  kl_NewObjRef ( lkg, lnk[0], obj[0], 19, r0 );
  kl_NewObjRef ( lkg, lnk[1], obj[0], 5, r1 );
  kl_NewObjRef ( lkg, lnk[2], obj[0], 8, r2 );
  kl_NewObjRef ( lkg, lnk[3], obj[0], 8, r3 );
  kl_NewObjRef ( lkg, lnk[4], obj[0], 8, r4 );
  kl_NewObjRef ( lkg, lnk[6], obj[0], 8, r5 );
  kl_NewObjRef ( lkg, lnk[7], obj[0], 8, r6 );
  kl_NewObjRef ( lkg, lnk[8], obj[0], 8, r7 );
  kl_NewObjRef ( lkg, lnk[10], obj[0], 8, r8 );
  kl_NewObjRef ( lkg, lnk[11], obj[0], 8, r9 );
  kl_NewObjRef ( lkg, lnk[12], obj[0], 8, r10 );
  kl_NewObjRef ( lkg, lnk[14], obj[0], 8, r11 );
  kl_NewObjRef ( lkg, lnk[15], obj[0], 8, r12 );
  kl_NewObjRef ( lkg, lnk[16], obj[0], 8, r13 );
  kl_NewObjRef ( lkg, lnk[18], obj[0], 8, r14 );
  kl_NewObjRef ( lkg, lnk[19], obj[0], 8, r15 );
  kl_NewObjRef ( lkg, lnk[20], obj[0], 8, r16 );
} /*ConstructPalmMeshObj*/

static void PrepareKLBuffers ( kl_linkage *lkg )
{
  GLuint *cpi;
  int    r, i, nv;

  nv = lkg->obj[0].nvert + 5*lkg->obj[1].nvert;
  if ( (cpi = malloc ( nv*sizeof(GLuint) )) ) {
    memset ( cpi, 0, nv*sizeof(GLuint) );
    for ( r = 0; r < lkg->norefs; r++ )
      if ( lkg->oref[r].on == 0 ) {
        for ( i = 0; i < lkg->oref[r].nv; i++ )
          cpi[lkg->oref[r].vn[i]] = r;
      }
    glGenBuffers ( 2, lktrbuf );
    glBindBuffer ( GL_SHADER_STORAGE_BUFFER, lktrbuf[0] );
    glBufferData ( GL_SHADER_STORAGE_BUFFER, lkg->norefs*16*sizeof(GLfloat),
                   NULL, GL_DYNAMIC_DRAW );
    glBindBuffer ( GL_SHADER_STORAGE_BUFFER, lktrbuf[1] );
    glBufferData ( GL_SHADER_STORAGE_BUFFER, nv*sizeof(GLuint),
                   cpi, GL_STATIC_DRAW );
    ExitIfGLError ( "PrepareKLBuffers" );
    free ( cpi );
  }
  else
    ExitOnError ( "PrepareKLBuffers" );
} /*PrepareKLBuffers*/

kl_linkage *ConstructPalmLinkage ( mypalmscene *scene )
{
  static int jtype[NKLJOINTS] =
    { KL_ART_ROT_Y, KL_ART_ROT_Z, KL_ART_ROT_X, KL_ART_ROT_X,
      KL_ART_ROT_Z, KL_ART_ROT_X, KL_ART_ROT_X, KL_ART_ROT_X,
      KL_ART_ROT_Z, KL_ART_ROT_X, KL_ART_ROT_X, KL_ART_ROT_X,
      KL_ART_ROT_Z, KL_ART_ROT_X, KL_ART_ROT_X, KL_ART_ROT_X,
      KL_ART_ROT_Z, KL_ART_ROT_X, KL_ART_ROT_X, KL_ART_ROT_X };
  static int jpnum[NKLARTPARAMS] =
    {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19};
  static float palmartp0[NKLARTPARAMS] =
      {0.0,0.0,0.0,0.0,0.2,0.0,0.0,0.0,0.5,0.0,0.0,
       0.0,0.5,0.0,0.0,0.0,0.8,0.0,0.0};
  static float pp0[3] = {-0.2, -0.86, 0.5},
               vp0[3] = {0.0,1.0,0.0}, vp1[3] = {0.22,0.83,0.1};

  kl_linkage *lkg;
  int        obj[1], lnk[NKLINKS], jnt[NKLJOINTS];
  GLfloat    tra[16];
  int        i, j, k, l;

  if ( (lkg = kl_NewLinkage ( NKLOBJ, NKLINKS, NKLREFS, NKLJOINTS,
                              NKLARTPARAMS, (void*)scene )) ) {
    memcpy ( scene->palmartp, palmartp0, NKLARTPARAMS*sizeof(float) );
    for ( i = 0; i < NKLINKS; i++ )
      lnk[i] = kl_NewLink ( lkg );
    ConstructPalmMeshObj ( scene, lkg, obj, lnk );
    for ( k = j = 0;  k < 5;  k++ ) {
      for ( i = 0, l = -1;  i < 4;  i++, l = j++ )
        jnt[j] = kl_NewJoint ( lkg, lnk[l+1], lnk[j+1], jtype[j], jpnum[j] );
    }
        /* kciuk */
    M4x4RotateP2Vf ( tra, pp0, vp0, vp1 );
    kl_SetJointFtr ( lkg, jnt[0], tra, true );
    M4x4Translatef ( tra, 0.125, -0.842, 0.0 );
    kl_SetJointFtr ( lkg, jnt[1], tra, true );
    M4x4Translatef ( tra, 0.329, -0.275, 0.05 );
    M4x4MRotateZf ( tra, -0.1*PI );
    M4x4MRotateYf ( tra, -0.2*PI );
    kl_SetJointFtr ( lkg, jnt[2], tra, true );
    M4x4Translatef ( tra, 0.4, -0.05, 0.05 );  
    M4x4MRotateZf ( tra, -0.1*PI );
    M4x4MRotateYf( tra, -0.2*PI );
    kl_SetJointFtr ( lkg, jnt[3], tra, true );
        /* wskazujacy */
    M4x4Translatef ( tra, 0.070, 0.05, 0.08 );
    kl_SetJointFtr ( lkg, jnt[4], tra, true );
    kl_SetJointFtr ( lkg, jnt[5], tra, true );
    M4x4Translatef ( tra, 0.094, 0.324, 0.08 );
    kl_SetJointFtr ( lkg, jnt[6], tra, true );
    M4x4Translatef ( tra, 0.1, 0.591, 0.08 );
    kl_SetJointFtr ( lkg, jnt[7], tra, true );
        /* srodkowy */
    M4x4Translatef ( tra, -0.13, 0.08, 0.08 );
    kl_SetJointFtr ( lkg, jnt[8], tra, true );
    kl_SetJointFtr ( lkg, jnt[9], tra, true );
    M4x4Translatef ( tra, -0.13, 0.349, 0.08 );
    kl_SetJointFtr ( lkg, jnt[10], tra, true );
    M4x4Translatef ( tra, -0.13, 0.634, 0.08 );
    kl_SetJointFtr ( lkg, jnt[11], tra, true );
        /* serdeczny */
    M4x4Translatef ( tra, -0.359, 0.05, 0.08 );
    kl_SetJointFtr ( lkg, jnt[12], tra, true );
    kl_SetJointFtr ( lkg, jnt[13], tra, true );
    M4x4Translatef ( tra, -0.374, 0.324, 0.08 );
    kl_SetJointFtr ( lkg, jnt[14], tra, true );
    M4x4Translatef ( tra, -0.392, 0.585, 0.08 );
    kl_SetJointFtr ( lkg, jnt[15], tra, true );
        /* maly */
    M4x4Translatef ( tra, -0.563, 0.0, 0.08 );
    kl_SetJointFtr ( lkg, jnt[16], tra, true );
    kl_SetJointFtr ( lkg, jnt[17], tra, true );
    M4x4Translatef ( tra, -0.59, 0.232, 0.08 );
    kl_SetJointFtr ( lkg, jnt[18], tra, true );
    M4x4Translatef ( tra, -0.605, 0.41, 0.08 );
    kl_SetJointFtr ( lkg, jnt[19], tra, true );
    PrepareKLBuffers ( lkg );
  }
  return scene->lkg = lkg;
} /*ConstructPalmLinkage*/

void SetArticulationParameter ( mypalmscene *scene, int pnum )
{
  float x, par;

  x = scene->palmartp[pnum];
  par = (1.0-x)*palmartprange[pnum][0] + x*palmartprange[pnum][1];
  kl_SetArtParam ( scene->lkg, pnum, 1, &par );
} /*SetArticulationParameter*/

void ArticulatePalmLinkage ( mypalmscene *scene )
{
  float   x;
  GLfloat par[NKLARTPARAMS];
  int     i;

  for ( i = 0; i < NKLARTPARAMS; i++ ) {
    x = scene->palmartp[i];
    par[i] = (1.0-x)*palmartprange[i][0] + x*palmartprange[i][1];
  }
  kl_SetArtParam ( scene->lkg, 0, NKLARTPARAMS, par );
  kl_Articulate ( scene->lkg );
} /*ArticulatePalmLinkage*/

