
#define MESHDEG         3
#define NPALMMESHES     4
#define NKLOBJ          2
#define NKLINKS        26
#define NKLREFS        27
#define NKLJOINTS      25
#define NKLARTPARAMS   21

#define MAXKEYFRAMES  100

typedef struct {
          float   left, right, top, bottom, near, far, rl, tb;
          int     prevx, prevy;
          float   viewer_rvec[4], viewer_rangle;
          GLfloat eyepos[4];
          GLfloat mm[16], mmti[16], vm[16], pm[16];
          char    animation;
          float   speed, model_rot_angle, model_rot_angle0;
        } myglwindata;

typedef struct {
          char            meshsw[NPALMMESHES+1];
          char            lod;
          float           palmartp[NKLARTPARAMS];
          GPUmesh         *palms[NPALMMESHES+2];
          BezierPatchObjf *fingernails[2];
          kl_linkage      *lkg;
          char            draw_shadows;
          char            bs_ok, qs_ok;
          int             lastkeyframe, lastbsknot;
          float           *keyknots, *keyparams, *qparams,
                          *bsknots, *bsparams, *qbsparams;
          char            animate_mm, animate_vp, animate_kl;
          double          time0, atime0;
        } mypalmscene;


extern GLuint trbuf;
extern GLint  trbofs[6];

extern myglwindata myglwdata;
extern mypalmscene myscene;
extern float       model_rot_axis[3], model_rot_angle0;
extern float       viewer_pos0[4];

extern const float palmartp0[NKLARTPARAMS];

void SetupModelMatrix ( myglwindata *wd, float axis[3], float angle );
void InitViewMatrix ( myglwindata *wd );
void SetupViewMatrix ( myglwindata *wd, float vk[3], float angk );
char RotateViewer ( myglwindata *wd, float width, float height,
                    int delta_xi, int delta_eta );

void LoadMyShaders ( void );
void InitMyObject ( void );
void DestroyMyObject ( void );
void DestroyMyShaders ( void );

void LoadRenderingShaders ( void );
void DestroyRenderingShaders ( void );
void DrawBezPatches ( BezierPatchObjf *bp, int lod, char final );
void DrawBezPNets ( BezierPatchObjf *bp, const GLfloat colour[4] );
void DrawMeshEdges ( GPUmesh *mesh, const GLfloat colour[4] );
void DrawMeshFacets ( GPUmesh *mesh, int mat, char nvectors, char final );

void DrawSceneToWindow ( mypalmscene *scene, myglwindata *wd, int w, int h );

void LoadLinkageArticulationShaders ( void );
void DestroyLinkageArticulationShaders ( void );

kl_linkage *ConstructPalmLinkage ( mypalmscene *scene );
void SetArticulationParameter ( mypalmscene *scene, int pnum );
void ArticulatePalmLinkage ( mypalmscene *scene );

char InitKeyFrames ( void );
void CleanupKeyFrames ( void );
void ToggleAnimation ( void );
void AdvanceAnimation ( void );
void FindKeyFrame ( char right );
void SetKeyFrame ( void );
void InsertKeyFrame ( void );
void ChangeKeyFrame ( void );
void DeleteKeyFrame ( void );
void KnotWidgetPoint ( int xi );

