
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "openglheader.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "utilities.h"
#include "quaternions.h"
#include "linkage.h"
#include "meshes.h"
#include "bezpatches.h"
#include "ispline.h"
#include "xvariables.h"
#include "xwidgets.h"
#include "xknotswidget.h"
#include "app3d_a.h"
#include "app3d_b.h"

void UpdateKeyInterpSpline ( void )
{
  myscene.bs_ok = ConstructCubicInterpBSplinef ( &myscene.lastbsknot,
                   myscene.bsknots, myscene.bsparams, myscene.lastkeyframe,
                   myscene.keyknots, NKLARTPARAMS, myscene.keyparams );
} /*UpdateKeyInterpSpline*/

void UpdateKeyInterpQSpline ( void )
{
  myscene.qs_ok = ConstructQuaternionInterpSplinef ( &myscene.lastbsknot,
                   myscene.bsknots, myscene.qbsparams, myscene.lastkeyframe,
                   myscene.keyknots, myscene.qparams );
} /*UpdateKeyInterpQSpline*/

char InitKeyFrames ( void  )
{
  int i;

  myscene.keyknots = malloc ( MAXKEYFRAMES*sizeof(float) );
  myscene.keyparams = malloc ( MAXKEYFRAMES*NKLARTPARAMS*sizeof(float) );
  myscene.bsknots = malloc ( (MAXKEYFRAMES+6)*sizeof(float) );
  myscene.bsparams = malloc ( (MAXKEYFRAMES+2)*NKLARTPARAMS*sizeof(float) );
  myscene.qparams = malloc ( MAXKEYFRAMES*4*sizeof(float) );
  myscene.qbsparams = malloc ( (MAXKEYFRAMES+2)*4*sizeof(float) );
  if ( myscene.keyknots && myscene.keyparams && myscene.qparams &&
       myscene.bsknots && myscene.bsparams && myscene.qbsparams ) {
    myscene.lastkeyframe = 3;
    myscene.keyknots[0] = 0.0;
    myscene.keyknots[1] = 10.0/3.0;
    myscene.keyknots[2] = 20.0/3.0;
    myscene.keyknots[3] = 10.0;
    for ( i = 0; i <= myscene.lastkeyframe; i++ )
      memcpy ( &myscene.keyparams[i*NKLARTPARAMS], palmartp0,
               NKLARTPARAMS*sizeof(float) );
    memset ( myscene.qparams, 0, (myscene.lastkeyframe+1)*4*sizeof(float) );
    for ( i = 0; i <= myscene.lastkeyframe; i++ )
      myscene.qparams[4*i] = 1.0;
    UpdateKeyInterpSpline ();
    UpdateKeyInterpQSpline ();
    myscene.animate_mm = myscene.animate_kl = true;
    myscene.animate_vp = false;
    return true;
  }
  else {
    CleanupKeyFrames ();
    return false;
  }
} /*InitKeyFrames*/

void CleanupKeyFrames ( void )
{
  if ( myscene.keyknots )  free ( myscene.keyknots );
  if ( myscene.keyparams ) free ( myscene.keyparams );
  if ( myscene.qparams )   free ( myscene.qparams );
  if ( myscene.bsknots )   free ( myscene.bsknots );
  if ( myscene.bsparams )  free ( myscene.bsparams );
  if ( myscene.qbsparams ) free ( myscene.qbsparams );
} /*CleanupKeyFrames*/

void FindKeyFrame ( char right )
{
  int   c;
  float a, v[3];

  c = FindKnotInterval ( myknotwdg->nknots, myscene.keyknots, myknotwdg->xc );
  if ( c < 0 )
    c = 0;
  else if ( c >= myknotwdg->nknots )
    c = myknotwdg->nknots-1;
  if ( !right && c > 0 && myknotwdg->xc == myscene.keyknots[c] )
    c --;
  else if ( right && c < myknotwdg->nknots-1 )
    c ++;
  myknotwdg->current = c;
  myknotwdg->curxi = KnotsWidgetXtoXi ( myknotwdg,
                                        myknotwdg->xc = myscene.keyknots[c] );
  memcpy ( myscene.palmartp, &myscene.keyparams[c*NKLARTPARAMS],
           NKLARTPARAMS*sizeof(float) );
  ArticulatePalmLinkage ( &myscene );
  RotVQuatf ( v, &a, &myscene.qparams[c*4] );
  SetupViewMatrix ( &myglwdata, v, a );
} /*FindKeyFrame*/

void SetKeyFrame ( void )
{
  int c;

  c = myknotwdg->current;
  if ( (c = myknotwdg->current) >= 0 && c < myknotwdg->nknots ) {
    memcpy ( &myscene.keyparams[c*NKLARTPARAMS], myscene.palmartp,
             NKLARTPARAMS*sizeof(float) );
    UpdateKeyInterpSpline ();
    QuatRotVf ( &myscene.qparams[c*4],
                myglwdata.viewer_rvec, myglwdata.viewer_rangle );
    UpdateKeyInterpQSpline ();
  }
} /*SetKeyFrame*/

void ClampArtParams ( float *params )
{
  int i;

  for ( i = 0; i < NKLARTPARAMS; i++ )
    if ( params[i] < 0.0 ) params[i] = 0.0;
    else if ( params[i] > 1.0 ) params[i] = 1.0;
} /*ClampArtParams*/

void InsertKeyFrame ( void )
{
  int c, n;

  myscene.lastkeyframe = myknotwdg->nknots-1;
  if ( (c = myknotwdg->current) < (n = myknotwdg->nknots)-1 ) {
    memmove ( &myscene.keyparams[(c+1)*NKLARTPARAMS],
              &myscene.keyparams[c*NKLARTPARAMS],
              (n-c-1)*NKLARTPARAMS*sizeof(float) );
    BSCdeBoorf ( 3, myscene.lastbsknot, myscene.bsknots, NKLARTPARAMS,
                 myscene.bsparams, myscene.keyknots[c], myscene.palmartp );
    ClampArtParams ( myscene.palmartp );
    memcpy ( &myscene.keyparams[c*NKLARTPARAMS], myscene.palmartp,
             NKLARTPARAMS*sizeof(float) );
    memmove ( &myscene.qparams[(c+1)*4], &myscene.qparams[c*4],
              (n-c-1)*4*sizeof(float) );
    if ( myscene.qs_ok )
      QuatSlerpdeBoorf ( 3, myscene.lastbsknot, myscene.bsknots, myscene.qbsparams,
                         myscene.keyknots[c], &myscene.qparams[c*4] );
  }
  else {
    memcpy ( &myscene.keyparams[c*NKLARTPARAMS],
             &myscene.keyparams[(c-1)*NKLARTPARAMS], NKLARTPARAMS*sizeof(float) );
    memcpy ( &myscene.qparams[c*4], &myscene.qparams[(c-1)*4], 4*sizeof(float) );
  }
  UpdateKeyInterpSpline ();
  ArticulatePalmLinkage ( &myscene );
  QuatRotVf ( &myscene.qparams[c*4],
              myglwdata.viewer_rvec, myglwdata.viewer_rangle );
  UpdateKeyInterpQSpline ();
} /*InsertKeyFrame*/

void SwapKeyFrames ( int i, int j )
{
  int   k, l, m;
  float *kp, a;

  kp = myscene.keyparams;
  for ( l = i*NKLARTPARAMS, m = j*NKLARTPARAMS, k = 0;  k < NKLARTPARAMS; k++ )
    { a = kp[l+k];  kp[l+k] = kp[m+k];  kp[m+k] = a; }
  kp = myscene.qparams;
  for ( l = i*4, m = j*4, k = 0;  k < 4; k++ )
    { a = kp[l+k];  kp[l+k] = kp[m+k];  kp[m+k] = a; }
} /*SwapKeyFrames*/

void ChangeKeyFrame ( void )
{
  int i, p, c;

  if ( (p = myknotwdg->prevc) < (c = myknotwdg->current) ) {
    for ( i = p; i < c; i++ )
      SwapKeyFrames ( i, i+1 );
  }
  else if ( p > c ) {
    for ( i = p; i > c; i-- )
      SwapKeyFrames ( i, i-1 );
  }
  UpdateKeyInterpSpline ();
  UpdateKeyInterpQSpline ();
} /*ChangeKeyFrame*/

void DeleteKeyFrame ( void )
{
  int c, n;

  myscene.lastkeyframe = myknotwdg->nknots-1;
  if ( (c = myknotwdg->current) < (n = myknotwdg->nknots)-1 ) {
    memmove ( &myscene.keyparams[c*NKLARTPARAMS],
              &myscene.keyparams[(c+1)*NKLARTPARAMS],
              (n-c-1)*NKLARTPARAMS*sizeof(float) );
    memmove ( &myscene.qparams[c*4], &myscene.qparams[(c+1)*4],
              (n-c-1)*4*sizeof(float) );
  }
   UpdateKeyInterpSpline ();
   UpdateKeyInterpQSpline ();
} /*DeleteKeyFrame*/

void ArticulateKLAtX ( float x )
{
  if ( x <= myscene.keyknots[0] || !myscene.bs_ok )
    memcpy ( myscene.palmartp, myscene.keyparams, NKLARTPARAMS*sizeof(float) );
  else if ( x >= myscene.keyknots[myscene.lastkeyframe] )
    memcpy ( myscene.palmartp,
             &myscene.keyparams[(myscene.lastkeyframe)*NKLARTPARAMS],
             NKLARTPARAMS*sizeof(float) );
  else {
    BSCdeBoorf ( 3, myscene.lastbsknot, myscene.bsknots, NKLARTPARAMS,
                 myscene.bsparams, x, myscene.palmartp );
    ClampArtParams ( myscene.palmartp );
  }
  ArticulatePalmLinkage ( &myscene );
} /*ArticulateKLAtX*/

void ArticulateVPosAtX ( float x )
{
  float q[4], v[3], a;

  if ( x <= myscene.keyknots[0] || !myscene.qs_ok )
    memcpy ( q, myscene.qparams, 4*sizeof(float) );
  else if ( x >= myscene.keyknots[myscene.lastkeyframe] )
    memcpy ( q, &myscene.qparams[4*myscene.lastkeyframe], 4*sizeof(float) );
  else
    QuatSlerpdeBoorf ( 3, myscene.lastbsknot, myscene.bsknots,
                       myscene.qbsparams, x, q );
  RotVQuatf ( v, &a, q );
  SetupViewMatrix ( &myglwdata, v, a );
} /*ArticulateVPosAtX*/

void KnotWidgetPoint ( int xi )
{
  float x;

  x = KnotsWidgetXitoX ( myknotwdg, xi );
  if ( myglwdata.animation ) {
    myscene.atime0 = app_time;
    myscene.time0 = x;
  }
  ArticulateKLAtX ( x );
  if ( myscene.animate_vp )
    ArticulateVPosAtX ( x );
} /*KnotWidgetPoint*/

void AdvanceAnimation ( void )
{
  double t;

  if ( myglwdata.animation ) {
    t = TimerToc ();
    if ( myscene.animate_mm ) {
      myglwdata.model_rot_angle = myglwdata.model_rot_angle0 + 0.78539816*t;
      SetupModelMatrix ( &myglwdata, model_rot_axis, myglwdata.model_rot_angle );
    }
    myknotwdg->xc = myscene.time0 + app_time - myscene.atime0;
    if ( myknotwdg->xc > myscene.keyknots[myknotwdg->nknots-1] )
      myknotwdg->xc = myscene.keyknots[myknotwdg->nknots-1];
    if ( myscene.animate_kl )
      ArticulateKLAtX ( myknotwdg->xc );
    if ( myscene.animate_vp )
      ArticulateVPosAtX ( myknotwdg->xc );
    wm1->changed = wm2->changed = wm3->changed = true;
    PostMenuExposeEvent ( wm1 );
    PostMenuExposeEvent ( wm2 );
    PostMenuExposeEvent ( wm3 );
    PostClientMessageEvent ( window[2], aAnimate, 8, NULL );
  }
} /*AdvanceAnimation*/

void ToggleAnimation ( void )
{
  if ( (myglwdata.animation = !myglwdata.animation) ) {
    TimerTic ();
    myscene.atime0 = app_time;
    if ( myknotwdg->xc >= myscene.keyknots[0] &&
         myknotwdg->xc < myscene.keyknots[myknotwdg->nknots-1] )
      myscene.time0 = myknotwdg->xc;
    else
      myscene.time0 = myknotwdg->xc = myscene.keyknots[0];
    myknotwdg->editswitch = myknotwdg->panswitch = false;
    PostClientMessageEvent ( window[2], aAnimate, 8, NULL );
  }
  else
    myglwdata.model_rot_angle0 = myglwdata.model_rot_angle;
  playbtn->state = myglwdata.animation ? WDGSTATE_DEFAULT+1 : WDGSTATE_DEFAULT;
/*  PostMenuExposeEvent ( wm1 );*/
  wm3->changed = true;
  PostMenuExposeEvent ( wm3 );
} /*ToggleAnimation*/

