
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "openglheader.h"

#include "utilities.h"
#include "meshes.h"

#define LOCAL_GROUP_SIZE 256
#define GRSIZE(a) (((a)+LOCAL_GROUP_SIZE-1)/LOCAL_GROUP_SIZE)

#define MVFBUF   mbuf[0]
#define MHEBUF   mbuf[1]
#define VCBUF    mbuf[2]

#define FHEMASK  0x01FFFFFF
#define TAGMASK  0x02000000
#define DEGMASK  0xFC000000
#define DEGSHIFT         26

#define SSB GL_SHADER_STORAGE_BUFFER

/* ////////////////////////////////////////////////////////////////////////// */
static GLuint nv_shader_id, nv_program_id;
static GLuint uvloc[9];

void LoadMeshNormalVectorShaders ( void )
{
  const GLchar *filename[] =
    { "mnv.glsl.comp" };
  const GLchar *uvname[] =
    { "innsattr", "pdim", "inpofs", "inv", "inhe", "infac",
      "outnsattr", "outpofs", "outnvofs" };
  int i;

  nv_shader_id = CompileShaderFiles ( GL_COMPUTE_SHADER, 1, &filename[0] );
  nv_program_id = LinkShaderProgram ( 1, &nv_shader_id );
  for ( i = 0; i < 9; i++ )
    uvloc[i] = glGetUniformLocation ( nv_program_id, uvname[i] );
  ExitIfGLError ( "LoadMeshNormalVectorShaders" );
} /*LoadMeshNormalVectorShaders*/

void DestroyMeshNormalVectorShaders ( void )
{
  glUseProgram ( 0 );
  glDeleteShader ( nv_shader_id );
  glDeleteProgram ( nv_program_id );
  ExitIfGLError ( "DestroyMeshNormalVectorShaders" );
} /*DestroyMeshNormalVectorShaders*/

void ComputeMeshNormalVectors ( GPUmesh *mesh, int nsattr, int nvofs )
{
  GLuint vcbuf;

  glGenBuffers ( 1, &vcbuf );
  glBindBufferBase ( SSB, 3, vcbuf );
  glBufferData ( SSB, mesh->nv*nsattr*sizeof(GLfloat), NULL, GL_DYNAMIC_DRAW );
  ExitIfGLError ( "ComputeMeshNormalVectors 0" );
  glUseProgram ( nv_program_id );
  glBindBufferBase ( SSB, 0, mesh->MVFBUF );
  glBindBufferBase ( SSB, 1, mesh->MHEBUF );
  glBindBufferBase ( SSB, 2, mesh->VCBUF );
  glUniform1i ( uvloc[0], mesh->nsattr );
  glUniform1i ( uvloc[1], mesh->pdim );
  glUniform1i ( uvloc[2], mesh->pofs );
  glUniform1i ( uvloc[3], mesh->nv );
  glUniform1i ( uvloc[4], mesh->nhe );
  glUniform1i ( uvloc[5], mesh->nfac );
  glUniform1i ( uvloc[6], nsattr );
  glUniform1i ( uvloc[7], mesh->pofs );
  glUniform1i ( uvloc[8], mesh->nvofs = nvofs );
  glDispatchCompute ( GRSIZE (mesh->nv), 1, 1 );
  ExitIfGLError ( "ComputeMeshNormalVectors 1" );
  glDeleteBuffers ( 1, &mesh->VCBUF );
  mesh->VCBUF = vcbuf;
  mesh->nsattr = nsattr;
  ExitIfGLError ( "ComputeMeshNormalVectors" );
} /*ComputeMeshNormalVectors*/

